

/********************************************/
/* Matrix.c - by Morgan Aldridge			*/
/********************************************/
/*   This source code is copyright (C) 2001 */
/* by Morgan Aldridge.  You are free to use */
/* and/or distibute this source code as long*/
/* as it remains in it's original form and  */
/* this message is included					*/
/********************************************/
/* If you find a bug in this code, then     */
/* e-mail classic@sover.net so that Morgan  */
/* Aldridge may fix any bugs that may remain*/
/********************************************/
/* This source code is not guaranteed to    */
/* work perfectly, so use it at your own    */
/* risk.  Morgan Aldridge will not be held  */
/* responsible to any damage which may      */
/* occur due to your use of this source code*/
/********************************************/


/**> HEADER FILES <**/
#include "Matrix.h"


/********************> Matrix_Identity() <*****/
void		Matrix_Identity( Matrix4 *matrix )
{
	
	int	temp = 0;
	
	/* reset the matrix to an identity matrix */
	matrix->m[INDEX(1,1)] = 1;
	matrix->m[INDEX(1,2)] = 0;
	matrix->m[INDEX(1,3)] = 0;
	matrix->m[INDEX(1,4)] = 0;
	matrix->m[INDEX(2,1)] = 0;
	matrix->m[INDEX(2,2)] = 1;
	matrix->m[INDEX(2,3)] = 0;
	matrix->m[INDEX(2,4)] = 0;
	matrix->m[INDEX(3,1)] = 0;
	matrix->m[INDEX(3,2)] = 0;
	matrix->m[INDEX(3,3)] = 1;
	matrix->m[INDEX(3,4)] = 0;
	matrix->m[INDEX(4,1)] = 0;
	matrix->m[INDEX(4,2)] = 0;
	matrix->m[INDEX(4,3)] = 0;
	matrix->m[INDEX(4,4)] = 1;
	
}

/********************> Matrix_Multiply() <*****/
Matrix4		Matrix_Multiply( Matrix4 matrix1, Matrix4 matrix2 )
{
	
	Matrix4	product;
	int	i, j, k;
	
	/* Multiply matrix1 by matrix2 */
	for ( i = 1; i <= 4; i++ )
		for ( j = 1; j <= 4; j++ )
			for ( k = 1, product.m[INDEX(i,j)] = 0; k <= 4; k++ )
				product.m[INDEX(i,j)] += matrix1.m[INDEX(i,k)] * matrix2.m[INDEX(k,j)];
	
	return product;
	
}

/********************> Matrix_Translate() <*****/
Matrix4		Matrix_Translate( Matrix4 matrix, double x, double y, double z )
{
	
	/* Multiply the matrix by the translation matrix */
	return Matrix_Multiply( matrix, Matrix_TranslationMatrix( x, y, z ) );
	
}

/********************> Matrix_Rotate() <*****/
Matrix4		Matrix_Rotate( Matrix4 matrix, double x, double y, double z )
{
	
	Matrix4	product;
	
	/* Set the product matrix equal to the matrix passed in */
	product = matrix;
	
	/* Multiply the product matrix by the x, y, and z rotation matrices */
	product = Matrix_Multiply( product, Matrix_XRotationMatrix( x ) );
	product = Matrix_Multiply( product, Matrix_YRotationMatrix( y ) );
	product = Matrix_Multiply( product, Matrix_ZRotationMatrix( z ) );
	
	return product;
	
}

/********************> Matrix_Scale() <*****/
Matrix4		Matrix_Scale( Matrix4 matrix, double sX, double sY, double sZ )
{
	
	/* Multiply the matrix by the scaling matrix */
	return Matrix_Multiply( matrix, Matrix_ScalingMatrix( sX, sY, sZ ) );
	
}

/********************> Matrix_MultiplyVertex() <*****/
Vector3D	Matrix_MultiplyVertex( Matrix4 matrix, double x, double y, double z )
{
	
	Vector3D	vertex;
	
	/* Multiply the vector (x, y, z) by the matrix */
	vertex.x = ( x * matrix.m[INDEX(1,1)] ) + ( y * matrix.m[INDEX(1,2)] ) + ( z * matrix.m[INDEX(1,3)] ) + matrix.m[INDEX(1,4)];
	vertex.y = ( x * matrix.m[INDEX(2,1)] ) + ( y * matrix.m[INDEX(2,2)] ) + ( z * matrix.m[INDEX(2,3)] ) + matrix.m[INDEX(2,4)];
	vertex.z = ( x * matrix.m[INDEX(3,1)] ) + ( y * matrix.m[INDEX(3,2)] ) + ( z * matrix.m[INDEX(3,3)] ) + matrix.m[INDEX(3,4)];
	
	return vertex;
	
}

/********************> Matrix_TranslationMatrix() <*****/
Matrix4		Matrix_TranslationMatrix( double x, double y, double z )
{
	
	Matrix4	matrix;
	
	/* Create a translation matrix */
	Matrix_Identity( &matrix );
	matrix.m[INDEX(1,4)] = x;
	matrix.m[INDEX(2,4)] = y;
	matrix.m[INDEX(3,4)] = z;
	
	return matrix;
	
}

/********************> Matrix_XRotationMatrix() <*****/
Matrix4		Matrix_XRotationMatrix( double x )
{
	
	Matrix4	matrix;
	
	/* Create an x rotation matrix */
	Matrix_Identity( &matrix );
	matrix.m[INDEX(2,2)] = cosf( x );
	matrix.m[INDEX(2,3)] = -sinf( x );
	matrix.m[INDEX(3,2)] = sinf( x );
	matrix.m[INDEX(3,3)] = cosf( x );
	
	return matrix;
	
}

/********************> Matrix_YRotationMatrix() <*****/
Matrix4		Matrix_YRotationMatrix( double y )
{
	
	Matrix4	matrix;
	
	/* Create a y rotation matrix */
	Matrix_Identity( &matrix );
	matrix.m[INDEX(1,1)] = cosf( y );
	matrix.m[INDEX(1,3)] = sinf( y );
	matrix.m[INDEX(3,1)] = -sinf( y );
	matrix.m[INDEX(3,3)] = cosf( y );
	
	return matrix;
	
}

/********************> Matrix_ZRotationMatrix() <*****/
Matrix4		Matrix_ZRotationMatrix( double z )
{
	
	Matrix4	matrix;
	
	/* Create a z rotation matrix */
	Matrix_Identity( &matrix );
	matrix.m[INDEX(1,1)] = cosf( z );
	matrix.m[INDEX(1,2)] = -sinf( z );
	matrix.m[INDEX(2,1)] = sinf( z );
	matrix.m[INDEX(2,2)] = cosf( z );
	
	return matrix;
	
}

/********************> Matrix_ScalingMatrix() <*****/
Matrix4		Matrix_ScalingMatrix( double sX, double sY, double sZ )
{
	
	Matrix4	matrix;
	
	/* Create a scaling matrix */
	Matrix_Identity( &matrix );
	matrix.m[INDEX(1,1)] = sX;
	matrix.m[INDEX(2,2)] = sY;
	matrix.m[INDEX(3,3)] = sZ;
	
	return matrix;
	
}

/********************> Matrix_Print() <*****/
void		Matrix_Print( Matrix4 matrix )
{
	
	int	i;
	
	printf( "Matrix Contents:\n" );
	
	/* Print the contents of the matrix */
	for ( i = 1; i <= 4; i++ )
	{
		printf ( "| %4.2f %4.2f %4.2f %4.2f |\n", matrix.m[INDEX(i,1)], matrix.m[INDEX(i,2)], matrix.m[INDEX(i,3)], matrix.m[INDEX(i,4)] );
	}
	
}
