#include "tyrlite.h"
#include "phong.h"

static dplane_t 	backplanes[MAX_MAP_PLANES];
static edgeshare_t	g_edgeshare[MAX_MAP_EDGES];
static vec3_t		g_face_centroids[MAX_MAP_EDGES];
vec_t				g_smoothing_threshold;

void MakeBackplanes()
{
    int             i;

    for (i = 0; i < numplanes; i++) {
        backplanes[i].dist = -dplanes[i].dist;
        VectorSubtract(vec3_origin, dplanes[i].normal, backplanes[i].normal);
    }
}

const dplane_t* getPlaneFromFace(const dface_t* const face)
{
    if (!face)
        Error("getPlaneFromFace() face was NULL\n");

    if (face->side)
        return &backplanes[face->planenum];
    else
        return &dplanes[face->planenum];
}

void PairEdges()
{
    int             i, j, k;
    dface_t*        f;
    edgeshare_t*    e;
	vec3_t 			centroid;// make centroids as we go
	vec_t			g_smoothing_angle = 50;// FIXME! - Make global switch
	
	g_smoothing_threshold = (vec_t)cos(DEGREES_TO_RADIANS(g_smoothing_angle));
	//printf ("g_smoothing_threshold = %4.3f\n",g_smoothing_threshold);
	//printf ("g_smoothing_angle = %4.3f\n",g_smoothing_angle);
	
    memset(&g_edgeshare, 0, sizeof(g_edgeshare));

    f = dfaces;
    for (i = 0; i < numfaces; i++, f++)
	{
    	VectorCopy(vec3_origin, centroid);

        for (j = 0; j < f->numedges; j++)
        {        	
            k = dsurfedges[f->firstedge + j];
            if (k < 0)
            {
                e = &g_edgeshare[-k];

                //hlassert(e->faces[1] == NULL);
                if(e->faces[1] != NULL) Error("Shared Edge: Face already assigned.");
                e->faces[1] = f;
                
                // Centroids
				VectorAdd(dvertexes[dedges[-k].v[1]].point, centroid, centroid);
				VectorAdd(dvertexes[dedges[-k].v[0]].point, centroid, centroid);
            }
            else
            {
                e = &g_edgeshare[k];

                //hlassert(e->faces[0] == NULL);
                if(e->faces[0] != NULL) Error("Shared Edge: Face already assigned.");
                e->faces[0] = f;
                
                // Centroids
                VectorAdd(dvertexes[dedges[k].v[0]].point, centroid, centroid);
                VectorAdd(dvertexes[dedges[k].v[1]].point, centroid, centroid);
            }
      		
      		// Centroids - Fixup for anything with an altered origin (rotating models/turrets mostly)
        	// Save them for moving direct lighting points towards the face center
        	VectorScale(centroid, 1.0 / (f->numedges * 2), centroid);
//        	VectorAdd(centroid, faceoffset[i], g_face_centroids[i]);
        	VectorCopy(centroid, g_face_centroids[i]);
        	
            if (e->faces[0] && e->faces[1])
            {
                // determine if coplanar
//                if (e->faces[0]->planenum == e->faces[1]->planenum)
//                    e->coplanar = true;
//                else
                {
                    // see if they fall into a "smoothing group" based on angle of the normals
                    vec3_t normals[2];

                    VectorCopy(getPlaneFromFace(e->faces[0])->normal, normals[0]);
                    VectorCopy(getPlaneFromFace(e->faces[1])->normal, normals[1]);

                    e->cos_normals_angle = DotProduct(normals[0], normals[1]);

                    if (e->cos_normals_angle > (1.0 - NORMAL_EPSILON))
                        e->coplanar = true;
                    else if (g_smoothing_threshold > 0.0)
                    {
                        if (e->cos_normals_angle >= g_smoothing_threshold)
                        {
                            VectorAdd(normals[0], normals[1], e->interface_normal);
                            VectorNormalize(e->interface_normal);
                        }
                    }
                }
            }
        }
    }
}

void GetPhongNormal(int facenum, vec3_t spot, vec3_t phongnormal)
{
    int             j;
    const dface_t*  f = dfaces + facenum;
    const dplane_t* p = getPlaneFromFace(f);
    vec3_t          facenormal;

    VectorCopy(p->normal, facenormal);
    VectorCopy(facenormal, phongnormal);

    if (g_smoothing_threshold > 0.0)
    {    	
        // Calculate modified point normal for surface
        // Use the edge normals if they are defined.  Bend the surface towards the edge normal(s)
        // Crude first attempt: find nearest edge normal and do a simple interpolation with facenormal.
        // Second attempt: find edge points+center that bound the point and do a three-point triangulation(baricentric)
        // Better third attempt: generate the point normals for all vertices and do baricentric triangulation.

        for (j = 0; j < f->numedges; j++)
        {
            vec3_t          p1;
            vec3_t          p2;
            vec3_t          v1;
            vec3_t          v2;
            vec3_t          vspot;
            unsigned        prev_edge;
            unsigned        next_edge;
            int             e;
            int             e1;
            int             e2;
            edgeshare_t*    es;
            edgeshare_t*    es1;
            edgeshare_t*    es2;
            float           a1;
            float           a2;
            float           aa;
            float           bb;
            float           ab;

            if (j)
                prev_edge = f->firstedge + ((j - 1) % f->numedges);
            else
                prev_edge = f->firstedge + f->numedges - 1;

            if ((j + 1) != f->numedges)
                next_edge = f->firstedge + ((j + 1) % f->numedges);
            else
                next_edge = f->firstedge;

            e = dsurfedges[f->firstedge + j];
            e1 = dsurfedges[prev_edge];
            e2 = dsurfedges[next_edge];

            es = &g_edgeshare[abs(e)];
            es1 = &g_edgeshare[abs(e1)];
            es2 = &g_edgeshare[abs(e2)];

            if ((es->coplanar && es1->coplanar && es2->coplanar) ||
                (VectorCompare(es->interface_normal, vec3_origin) &&
                 VectorCompare(es1->interface_normal, vec3_origin) &&
                 VectorCompare(es2->interface_normal, vec3_origin)))
                continue;

            if (e > 0) {
                VectorCopy(dvertexes[dedges[e].v[0]].point, p1);
                VectorCopy(dvertexes[dedges[e].v[1]].point, p2);
            } else {
                VectorCopy(dvertexes[dedges[-e].v[1]].point, p1);
                VectorCopy(dvertexes[dedges[-e].v[0]].point, p2);
            }

            // Adjust for origin-based models
            VectorAdd(p1, faceoffset[facenum], p1);
            VectorAdd(p2, faceoffset[facenum], p2);

            // Build vectors from the middle of the face to the edge vertexes and the sample pos.
            VectorSubtract(p1, g_face_centroids[facenum], v1);
            VectorSubtract(p2, g_face_centroids[facenum], v2);
            VectorSubtract(spot, g_face_centroids[facenum], vspot);

            aa = DotProduct(v1, v1);
            bb = DotProduct(v2, v2);
            ab = DotProduct(v1, v2);
            a1 = (bb * DotProduct(v1, vspot) - ab * DotProduct(vspot, v2)) / (aa * bb - ab * ab);
            a2 = (DotProduct(vspot, v2) - a1 * ab) / bb;

            // Test center to sample vector for inclusion between center to vertex vectors (Use dot product of vectors)
            if (a1 >= 0.0 && a2 >= 0.0)
            {
                // calculate distance from edge to pos
                vec3_t	n1, n2;
                vec3_t	temp;

                VectorAdd(es->interface_normal, es1->interface_normal, n1);

                if (VectorCompare(n1, vec3_origin))
                    VectorCopy(facenormal, n1);

                VectorNormalize(n1);

                VectorAdd(es->interface_normal, es2->interface_normal, n2);

                if (VectorCompare(n2, vec3_origin))
                    VectorCopy(facenormal, n2);

                VectorNormalize(n2);

                // Interpolate between the center and edge normals based on sample position
                VectorScale(facenormal, 1.0 - a1 - a2, phongnormal);
                VectorScale(n1, a1, temp);
                VectorAdd(phongnormal, temp, phongnormal);
                VectorScale(n2, a2, temp);
                VectorAdd(phongnormal, temp, phongnormal);
                VectorNormalize(phongnormal);
                break;
            }
        }
    }
}